@@ -1,6 +1,7 @@
 /*
     SPDX-FileCopyrightText: 2013 Mark Gaiser <markg85@gmail.com>
     SPDX-FileCopyrightText: 2016 Martin Klapetek <mklapetek@kde.org>
+    SPDX-FileCopyrightText: 2021 Carl Schwan <carlschwan@kde.org>
 
     SPDX-License-Identifier: GPL-2.0-or-later
 */
@@ -14,8 +15,10 @@
 #include <QDir>
 #include <QMetaObject>
 
+constexpr int maxEventDisplayed = 5;
+
 DaysModel::DaysModel(QObject *parent)
-    : QAbstractListModel(parent)
+    : QAbstractItemModel(parent)
     , m_pluginsManager(nullptr)
     , m_lastRequestedEventsStartDate(QDate())
     , m_agendaNeedsUpdate(false)
@@ -38,18 +41,38 @@ void DaysModel::setSourceData(QList<DayData> *data)
 
 int DaysModel::rowCount(const QModelIndex &parent) const
 {
-    Q_UNUSED(parent)
-    if (m_data->size() <= 0) {
-        return 0;
+    if (!parent.isValid()) {
+        // day count
+        if (m_data->size() <= 0) {
+            return 0;
+        } else {
+            return m_data->size();
+        }
     } else {
-        return m_data->size();
+        // event count
+        const auto &eventDatas = data(parent, Roles::Events).value<QList<CalendarEvents::EventData>>();
+        Q_ASSERT(eventDatas.count() <= maxEventDisplayed);
+        return eventDatas.count();
     }
 }
 
+int DaysModel::columnCount(const QModelIndex &parent) const
+{
+    Q_UNUSED(parent)
+    return 1;
+}
+
 QVariant DaysModel::data(const QModelIndex &index, int role) const
 {
-    if (index.isValid()) {
-        const DayData &currentData = m_data->at(index.row());
+    if (!index.isValid()) {
+        return {};
+    }
+
+    const int row = index.row();
+
+    if (!index.parent().isValid()) {
+        // Fetch days in month
+        const DayData &currentData = m_data->at(row);
         const QDate currentDate(currentData.yearNumber, currentData.monthNumber, currentData.dayNumber);
 
         switch (role) {
@@ -57,6 +80,10 @@ QVariant DaysModel::data(const QModelIndex &index, int role) const
             return currentData.isCurrent;
         case containsEventItems:
             return m_eventsData.contains(currentDate);
+        case Events:
+            return QVariant::fromValue(m_eventsData.values(currentDate));
+        case EventCount:
+            return m_eventsData.values(currentDate).count();
         case containsMajorEventItems:
             return hasMajorEventAtDate(currentDate);
         case containsMinorEventItems:
@@ -68,8 +95,20 @@ QVariant DaysModel::data(const QModelIndex &index, int role) const
         case yearNumber:
             return currentData.yearNumber;
         }
+    } else {
+        // Fetch event in day
+        const auto &eventDatas = data(index.parent(), Roles::Events).value<QList<CalendarEvents::EventData>>();
+        if (eventDatas.count() < row) {
+            return {};
+        }
+
+        const auto &eventData = eventDatas[row];
+        switch (role) {
+        case EventColor:
+            return eventData.eventColor();
+        }
     }
-    return QVariant();
+    return {};
 }
 
 void DaysModel::update()
@@ -78,7 +117,12 @@ void DaysModel::update()
         return;
     }
 
+    // We need to reset the model since m_data has already been changed here
+    // and we can't remove the events manually with beginRemoveRows() since
+    // we don't know where the old events were located.
+    beginResetModel();
     m_eventsData.clear();
+    endResetModel();
 
     const QDate modelFirstDay(m_data->at(0).yearNumber, m_data->at(0).monthNumber, m_data->at(0).dayNumber);
 
@@ -96,14 +140,38 @@ void DaysModel::update()
 void DaysModel::onDataReady(const QMultiHash<QDate, CalendarEvents::EventData> &data)
 {
     m_eventsData.reserve(m_eventsData.size() + data.size());
-    m_eventsData += data;
+    for (int i = 0; i < m_data->count(); i++) {
+        const DayData &currentData = m_data->at(i);
+        const QDate currentDate(currentData.yearNumber, currentData.monthNumber, currentData.dayNumber);
+        if (!data.values(currentDate).isEmpty()) {
+            // Make sure we don't display more than maxEventDisplayed events.
+            const int currentCount = m_eventsData.values(currentDate).count();
+            if (currentCount >= maxEventDisplayed) {
+                break;
+            }
+
+            const int addedEventCount = std::min(currentCount + data.values(currentDate).count(), maxEventDisplayed) - currentCount;
+
+            // Add event
+            beginInsertRows(index(i, 0), 0, addedEventCount - 1);
+            int stopCounter = currentCount;
+            for (const auto &dataDay : data.values(currentDate)) {
+                if (stopCounter >= maxEventDisplayed) {
+                    break;
+                }
+                stopCounter++;
+                m_eventsData.insert(currentDate, dataDay);
+            }
+            endInsertRows();
+        }
+    }
 
     if (data.contains(QDate::currentDate())) {
         m_agendaNeedsUpdate = true;
     }
 
     // only the containsEventItems roles may have changed
-    Q_EMIT dataChanged(index(0, 0), index(m_data->count() - 1, 0), {containsEventItems, containsMajorEventItems, containsMinorEventItems});
+    Q_EMIT dataChanged(index(0, 0), index(m_data->count() - 1, 0), {containsEventItems, containsMajorEventItems, containsMinorEventItems, Events, EventCount});
 
     Q_EMIT agendaUpdated(QDate::currentDate());
 }
@@ -125,10 +193,10 @@ void DaysModel::onEventModified(const CalendarEvents::EventData &data)
         m_agendaNeedsUpdate = true;
     }
 
-    for (const QDate date : qAsConst(updatesList)) {
+    for (const QDate date : std::as_const(updatesList)) {
         const QModelIndex changedIndex = indexForDate(date);
         if (changedIndex.isValid()) {
-            Q_EMIT dataChanged(changedIndex, changedIndex, {containsEventItems, containsMajorEventItems, containsMinorEventItems});
+            Q_EMIT dataChanged(changedIndex, changedIndex, {containsEventItems, containsMajorEventItems, containsMinorEventItems, EventColor});
         }
         Q_EMIT agendaUpdated(date);
     }
@@ -136,6 +204,12 @@ void DaysModel::onEventModified(const CalendarEvents::EventData &data)
 
 void DaysModel::onEventRemoved(const QString &uid)
 {
+    // HACK We should update the model with beginRemoveRows instead of
+    // using beginResetModel() since this creates a small visual glitches
+    // if an event is removed in Korganizer and the calendar is open.
+    // Using beginRemoveRows instead we make the code a lot more complex
+    // and if not done correctly will introduce bugs.
+    beginResetModel();
     QList<QDate> updatesList;
     auto i = m_eventsData.begin();
     while (i != m_eventsData.end()) {
@@ -151,13 +225,15 @@ void DaysModel::onEventRemoved(const QString &uid)
         m_agendaNeedsUpdate = true;
     }
 
-    for (const QDate date : qAsConst(updatesList)) {
+    for (const QDate date : std::as_const(updatesList)) {
         const QModelIndex changedIndex = indexForDate(date);
         if (changedIndex.isValid()) {
             Q_EMIT dataChanged(changedIndex, changedIndex, {containsEventItems, containsMajorEventItems, containsMinorEventItems});
         }
+
         Q_EMIT agendaUpdated(date);
     }
+    endResetModel();
 }
 
 QList<QObject *> DaysModel::eventsForDate(const QDate &date)
@@ -178,7 +254,7 @@ QList<QObject *> DaysModel::eventsForDate(const QDate &date)
         return b.type() > a.type() || b.startDateTime() > a.startDateTime();
     });
 
-    for (const CalendarEvents::EventData &event : qAsConst(events)) {
+    for (const CalendarEvents::EventData &event : std::as_const(events)) {
         m_qmlData << new EventDataDecorator(event, this);
     }
 
@@ -255,5 +331,26 @@ QHash<int, QByteArray> DaysModel::roleNames() const
             {containsMinorEventItems, "containsMinorEventItems"},
             {dayNumber, "dayNumber"},
             {monthNumber, "monthNumber"},
-            {yearNumber, "yearNumber"}};
+            {yearNumber, "yearNumber"},
+            {EventColor, "eventColor"},
+            {EventCount, "eventCount"},
+            {Events, "events"}};
 }
+
+QModelIndex DaysModel::index(int row, int column, const QModelIndex &parent) const
+{
+    if (parent.isValid()) {
+        return createIndex(row, column, (intptr_t)parent.row());
+    }
+    return createIndex(row, column, nullptr);
+}
+
+QModelIndex DaysModel::parent(const QModelIndex &child) const
+{
+    if (child.internalId()) {
+        return createIndex(child.internalId(), 0, nullptr);
+    }
+    return QModelIndex();
+}
+
+Q_DECLARE_METATYPE(CalendarEvents::EventData)
