@@ -1,58 +1,46 @@
 /*
- *   Copyright 2007 Matt Broadstone <mbroadst@gmail.com>
- *   Copyright 2007-2011 Aaron Seigo <aseigo@kde.org>
- *   Copyright 2007 Riccardo Iaconelli <riccardo@kde.org>
- *   Copyright (c) 2009 Chani Armitage <chani@kde.org>
- *
- *   This program is free software; you can redistribute it and/or modify
- *   it under the terms of the GNU Library General Public License as
- *   published by the Free Software Foundation; either version 2, or
- *   (at your option) any later version.
- *
- *   This program is distributed in the hope that it will be useful,
- *   but WITHOUT ANY WARRANTY; without even the implied warranty of
- *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *   GNU General Public License for more details
- *
- *   You should have received a copy of the GNU Library General Public
- *   License along with this program; if not, write to the
- *   Free Software Foundation, Inc.,
- *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
- */
+    SPDX-FileCopyrightText: 2007 Matt Broadstone <mbroadst@gmail.com>
+    SPDX-FileCopyrightText: 2007-2011 Aaron Seigo <aseigo@kde.org>
+    SPDX-FileCopyrightText: 2007 Riccardo Iaconelli <riccardo@kde.org>
+    SPDX-FileCopyrightText: 2009 Chani Armitage <chani@kde.org>
+
+    SPDX-License-Identifier: LGPL-2.0-or-later
+*/
 
 #include "corona.h"
 #include "private/corona_p.h"
 
+#include <QDebug>
 #include <QGuiApplication>
 #include <QMimeData>
 #include <QPainter>
+#include <QScreen>
 #include <QTimer>
-#include <QScreen>
-#include <QScreen>
+
+#include <KLocalizedString>
+#include <KWindowSystem>
 
 #include <cmath>
 
-#include <QDebug>
-#include <klocalizedstring.h>
-#include <kwindowsystem.h>
-
 #include "containment.h"
+#include "debug_p.h"
 #include "pluginloader.h"
-#include "packagestructure.h"
 #include "private/applet_p.h"
 #include "private/containment_p.h"
-#include "private/package_p.h"
 #include "private/timetracker.h"
-#include "debug_p.h"
+
+#if PLASMA_BUILD_DEPRECATED_SINCE(5, 83)
+#include "packagestructure.h"
+#include "private/package_p.h"
+#endif
 
 using namespace Plasma;
 
 namespace Plasma
 {
-
 Corona::Corona(QObject *parent)
-    : QObject(parent),
-      d(new CoronaPrivate(this))
+    : QObject(parent)
+    , d(new CoronaPrivate(this))
 {
 #ifndef NDEBUG
     // qCDebug(LOG_PLASMA) << "!!{} STARTUP TIME" << QTime().msecsTo(QTime::currentTime()) << "Corona ctor start";
@@ -74,18 +62,18 @@ Corona::~Corona()
     delete d;
 }
 
+#if PLASMA_BUILD_DEPRECATED_SINCE(5, 6)
 Plasma::Package Corona::package() const
 {
-    Package p;
-    p.d->internalPackage = new KPackage::Package(d->package);
-    return p;
+    return Package(d->package);
 }
 
 void Corona::setPackage(const Plasma::Package &package)
 {
     setKPackage(*package.d->internalPackage);
-    emit packageChanged(package);
+    Q_EMIT packageChanged(package);
 }
+#endif
 
 KPackage::Package Corona::kPackage() const
 {
@@ -95,7 +83,7 @@ KPackage::Package Corona::kPackage() const
 void Corona::setKPackage(const KPackage::Package &package)
 {
     d->package = package;
-    emit kPackageChanged(package);
+    Q_EMIT kPackageChanged(package);
 }
 
 void Corona::saveLayout(const QString &configName) const
@@ -113,32 +101,34 @@ void Corona::saveLayout(const QString &configName) const
 
 void Corona::exportLayout(KConfigGroup &config, QList<Containment *> containments)
 {
-    foreach (const QString &group, config.groupList()) {
+    const auto groupList = config.groupList();
+    for (const QString &group : groupList) {
         KConfigGroup cg(&config, group);
         cg.deleteGroup();
     }
 
-    //temporarily unlock so that removal works
+    // temporarily unlock so that removal works
     Types::ImmutabilityType oldImm = immutability();
     d->immutability = Types::Mutable;
 
     KConfigGroup dest(&config, "Containments");
     KConfigGroup dummy;
-    foreach (Plasma::Containment *c, containments) {
+    for (Plasma::Containment *c : std::as_const(containments)) {
         c->save(dummy);
         c->config().reparent(&dest);
 
-        //ensure the containment is unlocked
-        //this is done directly because we have to bypass any Types::SystemImmutable checks
+        // ensure the containment is unlocked
+        // this is done directly because we have to bypass any Types::SystemImmutable checks
         c->Applet::d->immutability = Types::Mutable;
-        foreach (Applet *a, c->applets()) {
+        const auto lstApplet = c->applets();
+        for (Applet *a : lstApplet) {
             a->d->immutability = Types::Mutable;
         }
 
         c->destroy();
     }
 
-    //restore immutability
+    // restore immutability
     d->immutability = oldImm;
 
     config.sync();
@@ -152,7 +142,7 @@ void Corona::requestConfigSync()
 
     // TODO: should we check into our immutability before doing this?
 
-    //NOTE: this is a pretty simplistic model: we simply save no more than CONFIG_SYNC_TIMEOUT
+    // NOTE: this is a pretty simplistic model: we simply save no more than CONFIG_SYNC_TIMEOUT
     //      after the first time this is called. not much of a heuristic for save points, but
     //      it should at least compress these activities a bit and provide a way for applet
     //      authors to ween themselves from the sync() disease. A more interesting/dynamic
@@ -171,7 +161,7 @@ void Corona::loadLayout(const QString &configName)
 {
     if (!configName.isEmpty() && configName != d->configName) {
         // if we have a new config name passed in, then use that as the config file for this Corona
-        d->config = 0;
+        d->config = nullptr;
         d->configName = configName;
     }
 
@@ -192,39 +182,91 @@ QList<Plasma::Containment *> Corona::importLayout(const KConfigGroup &conf)
     return d->importLayout(conf, true);
 }
 
+#if PLASMA_BUILD_DEPRECATED_SINCE(5, 46)
 Containment *Corona::containmentForScreen(int screen) const
 {
-    foreach (Containment *containment, d->containments) {
-        if (containment->screen() == screen &&
-                (containment->containmentType() == Plasma::Types::DesktopContainment ||
-                 containment->containmentType() == Plasma::Types::CustomContainment)) {
+    for (Containment *containment : std::as_const(d->containments)) {
+        if (containment->screen() == screen //
+            && (containment->containmentType() == Plasma::Types::DesktopContainment //
+                || containment->containmentType() == Plasma::Types::CustomContainment)) {
             return containment;
         }
     }
 
-    return 0;
+    return nullptr;
 }
 
-Containment *Corona::containmentForScreen(int screen,
-                                          const QString &defaultPluginIfNonExistent, const QVariantList &defaultArgs)
+Containment *Corona::containmentForScreen(int screen, const QString &defaultPluginIfNonExistent, const QVariantList &defaultArgs)
 {
-    Containment *containment = containmentForScreen(screen);
+    return containmentForScreen(screen, QString(), defaultPluginIfNonExistent, defaultArgs);
+}
+#endif
+
+Containment *Corona::containmentForScreen(int screen, const QString &activity, const QString &defaultPluginIfNonExistent, const QVariantList &defaultArgs)
+{
+    Containment *containment = nullptr;
+
+    for (Containment *cont : std::as_const(d->containments)) {
+        /* clang-format off */
+        if (cont->lastScreen() == screen
+            && (cont->activity().isEmpty() || cont->activity() == activity)
+            && (cont->containmentType() == Plasma::Types::DesktopContainment
+                || cont->containmentType() == Plasma::Types::CustomContainment)) { /* clang-format on */
+            containment = cont;
+        }
+    }
+
     if (!containment && !defaultPluginIfNonExistent.isEmpty()) {
         // screen requests are allowed to bypass immutability
         if (screen >= 0) {
             Plasma::Types::ImmutabilityType imm = d->immutability;
             d->immutability = Types::Mutable;
-            containment = d->addContainment(defaultPluginIfNonExistent, defaultArgs, 0, false);
-            if (containment) {
-               // containment->setScreen(screen);
-            }
+            containment = d->addContainment(defaultPluginIfNonExistent, defaultArgs, 0, screen, false);
+
             d->immutability = imm;
         }
     }
 
+    if (containment) {
+        containment->setActivity(activity);
+    }
     return containment;
 }
 
+QList<Containment *> Corona::containmentsForActivity(const QString &activity)
+{
+    QList<Containment *> conts;
+
+    if (activity.isEmpty()) {
+        return conts;
+    }
+
+    std::copy_if(d->containments.begin(), d->containments.end(), std::back_inserter(conts), [activity](Containment *cont) {
+        return cont->activity() == activity //
+            && (cont->containmentType() == Plasma::Types::DesktopContainment //
+                || cont->containmentType() == Plasma::Types::CustomContainment);
+    });
+
+    return conts;
+}
+
+QList<Containment *> Corona::containmentsForScreen(int screen)
+{
+    QList<Containment *> conts;
+
+    if (screen < 0) {
+        return conts;
+    }
+
+    std::copy_if(d->containments.begin(), d->containments.end(), std::back_inserter(conts), [screen](Containment *cont) {
+        return cont->lastScreen() == screen //
+            && (cont->containmentType() == Plasma::Types::DesktopContainment //
+                || cont->containmentType() == Plasma::Types::CustomContainment);
+    });
+
+    return conts;
+}
+
 QList<Containment *> Corona::containments() const
 {
     return d->containments;
@@ -247,19 +289,19 @@ KSharedConfigPtr Corona::config() const
 Containment *Corona::createContainment(const QString &name, const QVariantList &args)
 {
     if (d->immutability == Types::Mutable || args.contains(QVariant::fromValue(QStringLiteral("org.kde.plasma:force-create")))) {
-        return d->addContainment(name, args, 0);
+        return d->addContainment(name, args, 0, -1, false);
     }
 
-    return 0;
+    return nullptr;
 }
 
 Containment *Corona::createContainmentDelayed(const QString &name, const QVariantList &args)
 {
     if (d->immutability == Types::Mutable) {
-        return d->addContainment(name, args, 0, true);
+        return d->addContainment(name, args, 0, -1, true);
     }
 
-    return 0;
+    return nullptr;
 }
 
 int Corona::screenForContainment(const Containment *) const
@@ -284,7 +326,7 @@ QRect Corona::availableScreenRect(int id) const
 
 void Corona::loadDefaultLayout()
 {
-    //Default implementation does nothing
+    // Default implementation does nothing
 }
 
 Types::ImmutabilityType Corona::immutability() const
@@ -303,10 +345,10 @@ void Corona::setImmutability(const Types::ImmutabilityType immutable)
 #endif
     d->immutability = immutable;
     d->updateContainmentImmutability();
-    //tell non-containments that might care (like plasmaapp or a custom corona)
-    emit immutabilityChanged(immutable);
+    // tell non-containments that might care (like plasmaapp or a custom corona)
+    Q_EMIT immutabilityChanged(immutable);
 
-    //update our actions
+    // update our actions
     QAction *action = d->actions.action(QStringLiteral("lock widgets"));
     if (action) {
         if (d->immutability == Types::SystemImmutable) {
@@ -321,6 +363,25 @@ void Corona::setImmutability(const Types::ImmutabilityType immutable)
         }
     }
 
+    action = d->actions.action(QStringLiteral("edit mode"));
+    if (action) {
+        switch (d->immutability) {
+        case Types::UserImmutable:
+            action->setEnabled(false);
+            action->setVisible(true);
+            break;
+        case Types::SystemImmutable:
+            action->setEnabled(false);
+            action->setVisible(false);
+            break;
+        case Types::Mutable:
+        default:
+            action->setEnabled(true);
+            action->setVisible(true);
+            break;
+        }
+    }
+
     if (d->immutability != Types::SystemImmutable) {
         KConfigGroup cg(config(), "General");
 
@@ -329,17 +390,49 @@ void Corona::setImmutability(const Types::ImmutabilityType immutable)
         cg.writeEntry("immutability", (int)d->immutability);
         requestConfigSync();
     }
+
+    if (d->immutability != Types::Mutable) {
+        setEditMode(false);
+    }
+}
+
+void Corona::setEditMode(bool edit)
+{
+    if (edit == d->editMode || (edit && d->immutability != Plasma::Types::Mutable)) {
+        return;
+    }
+
+    QAction *editAction = d->actions.action(QStringLiteral("edit mode"));
+    if (editAction) {
+        if (edit) {
+            editAction->setText(i18n("Exit Edit Mode"));
+        } else {
+            editAction->setText(i18n("Enter Edit Mode"));
+        }
+    }
+
+    d->editMode = edit;
+    Q_EMIT editModeChanged(edit);
+}
+
+bool Corona::isEditMode() const
+{
+    return d->editMode;
 }
 
 QList<Plasma::Types::Location> Corona::freeEdges(int screen) const
 {
     QList<Plasma::Types::Location> freeEdges;
-    freeEdges << Plasma::Types::TopEdge << Plasma::Types::BottomEdge
-              << Plasma::Types::LeftEdge << Plasma::Types::RightEdge;
+    /* clang-format off */
+    freeEdges << Plasma::Types::TopEdge
+              << Plasma::Types::BottomEdge
+              << Plasma::Types::LeftEdge
+              << Plasma::Types::RightEdge;
+    /* clang-format on */
 
-    foreach (Containment *containment, containments()) {
-        if (containment->screen() == screen &&
-                freeEdges.contains(containment->location())) {
+    const auto containments = this->containments();
+    for (Containment *containment : containments) {
+        if (containment->screen() == screen && freeEdges.contains(containment->location())) {
             freeEdges.removeAll(containment->location());
         }
     }
@@ -353,15 +446,14 @@ KActionCollection *Corona::actions() const
 }
 
 CoronaPrivate::CoronaPrivate(Corona *corona)
-    : q(corona),
-      immutability(Types::Mutable),
-      config(0),
-      configSyncTimer(new QTimer(corona)),
-      actions(corona),
-      containmentsStarting(0)
+    : q(corona)
+    , immutability(Types::Mutable)
+    , config(nullptr)
+    , configSyncTimer(new QTimer(corona))
+    , actions(corona)
+    , containmentsStarting(0)
 {
-    //TODO: make Package path configurable
-    KConfigGroup config(KSharedConfig::openConfig(), "General");
+    // TODO: make Package path configurable
 
     if (QCoreApplication::instance()) {
         configName = QCoreApplication::instance()->applicationName() + QStringLiteral("-appletsrc");
@@ -382,7 +474,7 @@ void CoronaPrivate::init()
     configSyncTimer->setSingleShot(true);
     QObject::connect(configSyncTimer, SIGNAL(timeout()), q, SLOT(syncConfig()));
 
-    //some common actions
+    // some common actions
     actions.setConfigGroup(QStringLiteral("Shortcuts"));
 
     QAction *lockAction = actions.add<QAction>(QStringLiteral("lock widgets"));
@@ -394,9 +486,20 @@ void CoronaPrivate::init()
     lockAction->setShortcut(QKeySequence(QStringLiteral("alt+d, l")));
     lockAction->setShortcutContext(Qt::ApplicationShortcut);
 
-    //fake containment/applet actions
-    KActionCollection *containmentActions = AppletPrivate::defaultActions(q); //containment has to start with applet stuff
-    ContainmentPrivate::addDefaultActions(containmentActions); //now it's really containment
+    // fake containment/applet actions
+    KActionCollection *containmentActions = AppletPrivate::defaultActions(q); // containment has to start with applet stuff
+    ContainmentPrivate::addDefaultActions(containmentActions); // now it's really containment
+
+    QAction *editAction = actions.add<QAction>(QStringLiteral("edit mode"));
+    QObject::connect(editAction, &QAction::triggered, q, [this]() {
+        q->setEditMode(!q->isEditMode());
+    });
+    editAction->setText(i18n("Enter Edit Mode"));
+    editAction->setAutoRepeat(true);
+    editAction->setIcon(QIcon::fromTheme(QStringLiteral("document-edit")));
+    editAction->setData(Plasma::Types::ControlAction);
+    editAction->setShortcut(QKeySequence(QStringLiteral("alt+d, e")));
+    editAction->setShortcutContext(Qt::ApplicationShortcut);
 }
 
 void CoronaPrivate::toggleImmutability()
@@ -411,7 +514,7 @@ void CoronaPrivate::toggleImmutability()
 void CoronaPrivate::saveLayout(KSharedConfigPtr cg) const
 {
     KConfigGroup containmentsGroup(cg, "Containments");
-    foreach (const Containment *containment, containments) {
+    for (const Containment *containment : containments) {
         QString cid = QString::number(containment->id());
         KConfigGroup containmentConfig(&containmentsGroup, cid);
         containment->save(containmentConfig);
@@ -420,7 +523,7 @@ void CoronaPrivate::saveLayout(KSharedConfigPtr cg) const
 
 void CoronaPrivate::updateContainmentImmutability()
 {
-    foreach (Containment *c, containments) {
+    for (Containment *c : std::as_const(containments)) {
         // we need to tell each containment that immutability has been altered
         c->updateConstraints(Types::ImmutableConstraint);
     }
@@ -444,14 +547,14 @@ void CoronaPrivate::containmentDestroyed(QObject *obj)
 void CoronaPrivate::syncConfig()
 {
     q->config()->sync();
-    emit q->configSynced();
+    Q_EMIT q->configSynced();
 }
 
-Containment *CoronaPrivate::addContainment(const QString &name, const QVariantList &args, uint id, bool delayedInit)
+Containment *CoronaPrivate::addContainment(const QString &name, const QVariantList &args, uint id, int lastScreen, bool delayedInit)
 {
     QString pluginName = name;
-    Containment *containment = 0;
-    Applet *applet = 0;
+    Containment *containment = nullptr;
+    Applet *applet = nullptr;
 
     // qCDebug(LOG_PLASMA) << "Loading" << name << args << id;
 
@@ -484,8 +587,11 @@ Containment *CoronaPrivate::addContainment(const QString &name, const QVariantLi
             applet->init();
             delete applet;
         }
-        applet = containment = new Containment(q, {}, id);
-        //if it's a dummy containment, just say its ui is ready, not blocking the corona
+        applet = containment = new Containment(q, KPluginMetaData(), QVariantList{QVariant(), QVariant(), id});
+        if (lastScreen >= 0) {
+            containment->d->lastScreen = lastScreen;
+        }
+        // if it's a dummy containment, just say its ui is ready, not blocking the corona
         applet->updateConstraints(Plasma::Types::UiReadyConstraint);
 
         // we want to provide something and don't care about the failure to launch
@@ -500,18 +606,15 @@ Containment *CoronaPrivate::addContainment(const QString &name, const QVariantLi
         conf.deleteGroup();
     }
 
-    //make sure the containments are sorted by id
-    auto position = std::lower_bound(containments.begin(), containments.end(), containment, [](Plasma::Containment *c1,  Plasma::Containment *c2) {
+    // make sure the containments are sorted by id
+    auto position = std::lower_bound(containments.begin(), containments.end(), containment, [](Plasma::Containment *c1, Plasma::Containment *c2) {
         return c1->id() < c2->id();
     });
     containments.insert(position, containment);
 
-    QObject::connect(containment, SIGNAL(destroyed(QObject*)),
-                     q, SLOT(containmentDestroyed(QObject*)));
-    QObject::connect(containment, SIGNAL(configNeedsSaving()),
-                     q, SLOT(requestConfigSync()));
-    QObject::connect(containment, SIGNAL(screenChanged(int)),
-                     q, SIGNAL(screenOwnerChanged(int)));
+    QObject::connect(containment, SIGNAL(destroyed(QObject *)), q, SLOT(containmentDestroyed(QObject *)));
+    QObject::connect(containment, &Applet::configNeedsSaving, q, &Corona::requestConfigSync);
+    QObject::connect(containment, &Containment::screenChanged, q, &Corona::screenOwnerChanged);
 
     if (!delayedInit) {
         containment->init();
@@ -521,10 +624,10 @@ Containment *CoronaPrivate::addContainment(const QString &name, const QVariantLi
         containment->save(cg);
         q->requestConfigSync();
         containment->flushPendingConstraintsEvents();
-        emit q->containmentAdded(containment);
-        //if id = 0 a new containment has been created, not restored
+        Q_EMIT q->containmentAdded(containment);
+        // if id = 0 a new containment has been created, not restored
         if (id == 0) {
-            emit q->containmentCreated(containment);
+            Q_EMIT q->containmentCreated(containment);
         }
     }
 
@@ -540,15 +643,15 @@ QList<Plasma::Containment *> CoronaPrivate::importLayout(const KConfigGroup &con
     QList<Plasma::Containment *> newContainments;
     QSet<uint> containmentsIds;
 
-    foreach (Containment *containment, containments) {
+    for (Containment *containment : std::as_const(containments)) {
         containmentsIds.insert(containment->id());
     }
 
     KConfigGroup containmentsGroup(&conf, "Containments");
     QStringList groups = containmentsGroup.groupList();
-    qSort(groups.begin(), groups.end());
+    std::sort(groups.begin(), groups.end());
 
-    foreach (const QString &group, groups) {
+    for (const QString &group : std::as_const(groups)) {
         KConfigGroup containmentConfig(&containmentsGroup, group);
 
         if (containmentConfig.entryMap().isEmpty()) {
@@ -570,11 +673,12 @@ QList<Plasma::Containment *> CoronaPrivate::importLayout(const KConfigGroup &con
             containmentConfig.copyTo(&realConf);
         }
 
-        //qCDebug(LOG_PLASMA) << "got a containment in the config, trying to make a" << containmentConfig.readEntry("plugin", QString()) << "from" << group;
+        // qCDebug(LOG_PLASMA) << "got a containment in the config, trying to make a" << containmentConfig.readEntry("plugin", QString()) << "from" << group;
 #ifndef NDEBUG
-        // qCDebug(LOG_PLASMA) << "!!{} STARTUP TIME" << QTime().msecsTo(QTime::currentTime()) << "Adding Containment" << containmentConfig.readEntry("plugin", QString());
+        // qCDebug(LOG_PLASMA) << "!!{} STARTUP TIME" << QTime().msecsTo(QTime::currentTime()) << "Adding Containment" << containmentConfig.readEntry("plugin",
+        // QString());
 #endif
-        Containment *c = addContainment(containmentConfig.readEntry("plugin", QString()), QVariantList(), cid);
+        Containment *c = addContainment(containmentConfig.readEntry("plugin", QString()), QVariantList(), cid, -1);
         if (!c) {
             continue;
         }
@@ -597,15 +701,17 @@ QList<Plasma::Containment *> CoronaPrivate::importLayout(const KConfigGroup &con
 void CoronaPrivate::notifyContainmentsReady()
 {
     containmentsStarting = 0;
-    foreach (Containment *containment, containments) {
+    for (Containment *containment : std::as_const(containments)) {
         if (!containment->isUiReady() && containment->screen() >= 0) {
             ++containmentsStarting;
-            QObject::connect(containment, &Plasma::Containment::uiReadyChanged, q, [this](bool ready) { containmentReady(ready); } );
+            QObject::connect(containment, &Plasma::Containment::uiReadyChanged, q, [this](bool ready) {
+                containmentReady(ready);
+            });
         }
     }
 
     if (containmentsStarting <= 0) {
-        emit q->startupCompleted();
+        Q_EMIT q->startupCompleted();
     }
 }
 
@@ -616,7 +722,7 @@ void CoronaPrivate::containmentReady(bool ready)
     }
     --containmentsStarting;
     if (containmentsStarting <= 0) {
-        emit q->startupCompleted();
+        Q_EMIT q->startupCompleted();
     }
 }
 
