@@ -9,7 +9,6 @@
 #include <QPointer>
 #include <QStandardPaths>
 
-#include <KPluginLoader>
 #include <KService>
 #include <KServiceTypeTrader>
 #include <QDebug>
@@ -68,10 +67,10 @@ public:
         // anyway
         int maxCacheAge = 20;
         qint64 pluginCacheAge = 0;
-        QHash<QString, QVector<KPluginMetaData>> plugins;
+        QHash<QString, KPluginMetaData> plugins;
 
     public:
-        QVector<KPluginMetaData> findPluginsById(const QString &name, const QStringList &dirs);
+        KPluginMetaData findPluginById(const QString &name, const QString &pluginNamespace);
     };
     Cache plasmoidCache;
     Cache dataengineCache;
@@ -126,23 +125,21 @@ PluginLoader::PluginLoader()
 PluginLoader::~PluginLoader()
 {
 #if PLASMA_BUILD_DEPRECATED_SINCE(5, 83)
-    for (const auto &wp : qAsConst(d->structures)) {
+    for (const auto &wp : std::as_const(d->structures)) {
         delete wp;
     }
 #endif
     delete d;
 }
 
+#if PLASMA_BUILD_DEPRECATED_SINCE(5, 86)
 void PluginLoader::setPluginLoader(PluginLoader *loader)
 {
     if (!s_pluginLoader) {
         s_pluginLoader = loader;
-    } else {
-#ifndef NDEBUG
-        // qCDebug(LOG_PLASMA) << "Cannot set pluginLoader, already set!" << s_pluginLoader;
-#endif
     }
 }
+#endif
 
 PluginLoader *PluginLoader::self()
 {
@@ -163,36 +160,55 @@ Applet *PluginLoader::loadApplet(const QString &name, uint appletId, const QVari
         return nullptr;
     }
 
+#if PLASMA_BUILD_DEPRECATED_SINCE(5, 86)
     Applet *applet = d->isDefaultLoader ? nullptr : internalLoadApplet(name, appletId, args);
     if (applet) {
         return applet;
     }
+#else
+    Applet *applet = nullptr;
+#endif
 
     if (appletId == 0) {
         appletId = ++AppletPrivate::s_maxAppletId;
     }
 
     // Need to pass the empty directory because it's where plasmoids used to be
-    auto plugins = d->plasmoidCache.findPluginsById(name, {PluginLoaderPrivate::s_plasmoidsPluginDir, {}});
+    auto plugin = d->plasmoidCache.findPluginById(name, PluginLoaderPrivate::s_plasmoidsPluginDir);
+#if PLASMA_BUILD_DEPRECATED_SINCE(5, 86)
+    if (!plugin.isValid()) {
+        auto plugin = d->plasmoidCache.findPluginById(name, QString());
+        if (plugin.isValid()) {
+            qCWarning(LOG_PLASMA) << "The applet" << plugin.fileName()
+                                  << "is installed in a deprecated location, install it in the plasma/applets namespace instead";
+        }
+    }
+#endif
 
     const KPackage::Package p = KPackage::PackageLoader::self()->loadPackage(QStringLiteral("Plasma/Applet"), name);
 
     // If the applet is using another applet package, search for the plugin of the other applet
-    if (plugins.isEmpty()) {
+    if (!plugin.isValid()) {
         const QString parentPlugin = p.metadata().value(QStringLiteral("X-Plasma-RootPath"));
         if (!parentPlugin.isEmpty()) {
-            plugins = d->plasmoidCache.findPluginsById(parentPlugin, {PluginLoaderPrivate::s_plasmoidsPluginDir, {}});
+            plugin = d->plasmoidCache.findPluginById(parentPlugin, PluginLoaderPrivate::s_plasmoidsPluginDir);
+#if PLASMA_BUILD_DEPRECATED_SINCE(5, 86)
+            if (!plugin.isValid()) {
+                auto plugin = d->plasmoidCache.findPluginById(name, QString());
+                if (plugin.isValid()) {
+                    qCWarning(LOG_PLASMA) << "The applet" << plugin.fileName()
+                                          << "is installed in a deprecated location, install it in the plasma/applets namespace instead";
+                }
+            }
+#endif
         }
     }
 
-    if (!plugins.isEmpty()) {
-        KPluginLoader loader(plugins.first().fileName());
-        KPluginFactory *factory = loader.factory();
-        if (factory) {
-            QVariantList allArgs;
-            allArgs << QVariant::fromValue(p) << loader.metaData().toVariantMap() << appletId << args;
-            applet = factory->create<Plasma::Applet>(nullptr, allArgs);
-        }
+    if (plugin.isValid()) {
+        QPluginLoader loader(plugin.fileName());
+        QVariantList allArgs = {QVariant::fromValue(p), loader.metaData().toVariantMap(), appletId};
+        allArgs << args;
+        applet = KPluginFactory::instantiatePlugin<Plasma::Applet>(plugin, nullptr, allArgs).plugin;
     }
     if (applet) {
         return applet;
@@ -205,9 +221,9 @@ Applet *PluginLoader::loadApplet(const QString &name, uint appletId, const QVari
         allArgs << QVariant::fromValue(p) << p.metadata().fileName() << appletId << args;
 
         if (p.metadata().serviceTypes().contains(QLatin1String("Plasma/Containment"))) {
-            applet = new Containment(nullptr, allArgs);
+            applet = new Containment(nullptr, p.metadata(), allArgs);
         } else {
-            applet = new Applet(nullptr, allArgs);
+            applet = new Applet(nullptr, p.metadata(), allArgs);
         }
     }
 
@@ -220,18 +236,20 @@ Applet *PluginLoader::loadApplet(const QString &name, uint appletId, const QVari
 
 DataEngine *PluginLoader::loadDataEngine(const QString &name)
 {
+#if PLASMA_BUILD_DEPRECATED_SINCE(5, 86)
     DataEngine *engine = d->isDefaultLoader ? nullptr : internalLoadDataEngine(name);
     if (engine) {
         return engine;
     }
+#else
+    DataEngine *engine = nullptr;
+#endif
 
     // Look for C++ plugins first
-    const QVector<KPluginMetaData> plugins = d->dataengineCache.findPluginsById(name, {PluginLoaderPrivate::s_dataEnginePluginDir});
-    if (!plugins.isEmpty()) {
-        KPluginLoader loader(plugins.constFirst().fileName());
-        const QVariantList argsWithMetaData = QVariantList() << loader.metaData().toVariantMap();
-        KPluginFactory *factory = loader.factory();
-        return factory ? factory->create<Plasma::DataEngine>(nullptr, argsWithMetaData) : nullptr;
+    KPluginMetaData plugin = d->dataengineCache.findPluginById(name, PluginLoaderPrivate::s_dataEnginePluginDir);
+    if (plugin.isValid()) {
+        const QVariantList args{QPluginLoader(plugin.fileName()).metaData().toVariantMap()};
+        engine = KPluginFactory::instantiatePlugin<Plasma::DataEngine>(plugin, nullptr, args).plugin;
     }
     if (engine) {
         return engine;
@@ -254,12 +272,12 @@ QStringList PluginLoader::listAllEngines(const QString &parentApp)
     };
     QVector<KPluginMetaData> plugins;
     if (parentApp.isEmpty()) {
-        plugins = KPluginLoader::findPlugins(PluginLoaderPrivate::s_dataEnginePluginDir);
+        plugins = KPluginMetaData::findPlugins(PluginLoaderPrivate::s_dataEnginePluginDir);
     } else {
-        plugins = KPluginLoader::findPlugins(PluginLoaderPrivate::s_dataEnginePluginDir, filter);
+        plugins = KPluginMetaData::findPlugins(PluginLoaderPrivate::s_dataEnginePluginDir, filter);
     }
 
-    for (auto &plugin : qAsConst(plugins)) {
+    for (auto &plugin : std::as_const(plugins)) {
         engines << plugin.pluginId();
     }
 
@@ -293,9 +311,9 @@ KPluginInfo::List PluginLoader::listEngineInfoByCategory(const QString &category
     };
     QVector<KPluginMetaData> plugins;
     if (parentApp.isEmpty()) {
-        plugins = KPluginLoader::findPlugins(PluginLoaderPrivate::s_dataEnginePluginDir, filterNormal);
+        plugins = KPluginMetaData::findPlugins(PluginLoaderPrivate::s_dataEnginePluginDir, filterNormal);
     } else {
-        plugins = KPluginLoader::findPlugins(PluginLoaderPrivate::s_dataEnginePluginDir, filterParentApp);
+        plugins = KPluginMetaData::findPlugins(PluginLoaderPrivate::s_dataEnginePluginDir, filterParentApp);
     }
 
     list = KPluginInfo::fromMetaData(plugins);
@@ -310,10 +328,14 @@ KPluginInfo::List PluginLoader::listEngineInfoByCategory(const QString &category
 
 Service *PluginLoader::loadService(const QString &name, const QVariantList &args, QObject *parent)
 {
+#if PLASMA_BUILD_DEPRECATED_SINCE(5, 86)
     Service *service = d->isDefaultLoader ? nullptr : internalLoadService(name, args, parent);
     if (service) {
         return service;
     }
+#else
+    Service *service = nullptr;
+#endif
 
     // TODO: scripting API support
     if (name.isEmpty()) {
@@ -323,17 +345,9 @@ Service *PluginLoader::loadService(const QString &name, const QVariantList &args
     }
 
     // Look for C++ plugins first
-    auto filter = [&name](const KPluginMetaData &md) -> bool {
-        return md.pluginId() == name;
-    };
-    QVector<KPluginMetaData> plugins = KPluginLoader::findPlugins(PluginLoaderPrivate::s_servicesPluginDir, filter);
-
-    if (!plugins.isEmpty()) {
-        KPluginLoader loader(plugins.first().fileName());
-        KPluginFactory *factory = loader.factory();
-        if (factory) {
-            service = factory->create<Plasma::Service>(nullptr, args);
-        }
+    KPluginMetaData plugin = KPluginMetaData::findPluginById(PluginLoaderPrivate::s_servicesPluginDir, name);
+    if (plugin.isValid()) {
+        service = KPluginFactory::instantiatePlugin<Plasma::Service>(plugin, parent, args).plugin;
     }
 
     if (service) {
@@ -351,24 +365,20 @@ ContainmentActions *PluginLoader::loadContainmentActions(Containment *parent, co
     if (name.isEmpty()) {
         return nullptr;
     }
-
+#if PLASMA_BUILD_DEPRECATED_SINCE(5, 86)
     ContainmentActions *actions = d->isDefaultLoader ? nullptr : internalLoadContainmentActions(parent, name, args);
     if (actions) {
         return actions;
     }
+#endif
 
-    const QVector<KPluginMetaData> plugins = d->containmentactionCache.findPluginsById(name, {PluginLoaderPrivate::s_containmentActionsPluginDir});
+    KPluginMetaData plugin = d->containmentactionCache.findPluginById(name, PluginLoaderPrivate::s_containmentActionsPluginDir);
 
-    if (!plugins.isEmpty()) {
-        KPluginLoader loader(plugins.first().fileName());
-        KPluginFactory *factory = loader.factory();
-        if (factory) {
-            actions = factory->create<Plasma::ContainmentActions>(nullptr, {QVariant::fromValue(plugins.first())});
+    if (plugin.isValid()) {
+        if (auto res = KPluginFactory::instantiatePlugin<Plasma::ContainmentActions>(plugin, nullptr, {QVariant::fromValue(plugin)})) {
+            return res.plugin;
         }
     }
-    if (actions) {
-        return actions;
-    }
 
     // FIXME: this is only for backwards compatibility, but probably will have to stay
     // for the time being
@@ -383,20 +393,12 @@ ContainmentActions *PluginLoader::loadContainmentActions(Containment *parent, co
     }
 
     KService::Ptr offer = offers.first();
-    KPluginLoader plugin(*offer);
 
+    KPluginMetaData data(offer->library());
     QVariantList allArgs;
     allArgs << offer->storageId() << args;
-    QString error;
-    actions = offer->createInstance<Plasma::ContainmentActions>(parent, allArgs, &error);
 
-    if (!actions) {
-#ifndef NDEBUG
-        // qCDebug(LOG_PLASMA) << "Couldn't load containmentActions \"" << name << "\"! reason given: " << error;
-#endif
-    }
-
-    return actions;
+    return KPluginFactory::instantiatePlugin<Plasma::ContainmentActions>(data, nullptr, allArgs).plugin;
 }
 
 #if PLASMA_BUILD_DEPRECATED_SINCE(5, 83)
@@ -431,15 +433,13 @@ Package PluginLoader::loadPackage(const QString &packageFormat, const QString &s
             return md.value(QStringLiteral("X-KDE-PluginInfo-Name")) == packageFormat;
         };
 
-        const QVector<KPluginMetaData> plugins = KPluginLoader::findPlugins(PluginLoaderPrivate::s_packageStructurePluginDir, filter);
+        const QVector<KPluginMetaData> plugins = KPluginMetaData::findPlugins(PluginLoaderPrivate::s_packageStructurePluginDir, filter);
 
         if (!plugins.isEmpty()) {
-            KPluginLoader loader(plugins.first().fileName());
-            KPluginFactory *factory = loader.factory();
-            if (!factory) {
-                qWarning() << "Error loading plugin:" << loader.errorString();
+            if (auto res = KPluginFactory::instantiatePlugin<Plasma::PackageStructure>(plugins.first())) {
+                structure = res.plugin;
             } else {
-                structure = factory->create<Plasma::PackageStructure>();
+                qWarning() << "Error loading plugin:" << res.errorString;
             }
 
             if (structure) {
@@ -453,10 +453,6 @@ Package PluginLoader::loadPackage(const QString &packageFormat, const QString &s
         return Package(structure);
     }
 
-#ifndef NDEBUG
-    // qCDebug(LOG_PLASMA) << "Couldn't load Package for" << packageFormat << "! reason given: " << error;
-#endif
-
     return Package();
 }
 #endif
@@ -520,10 +516,6 @@ QList<KPluginMetaData> PluginLoader::listAppletMetaData(const QString &category,
         };
     }
 
-    QList<KPluginMetaData> list;
-    if (!d->isDefaultLoader && (parentApp.isEmpty() || parentApp == QCoreApplication::instance()->applicationName())) {
-        list = KPluginInfo::toMetaData(internalAppletInfo(category)).toList();
-    }
     return KPackage::PackageLoader::self()->findPackages(QStringLiteral("Plasma/Applet"), QString(), filter);
 }
 
@@ -591,9 +583,6 @@ QList<KPluginMetaData> PluginLoader::listAppletMetaDataForUrl(const QUrl &url)
             QRegExp rx(glob);
             rx.setPatternSyntax(QRegExp::Wildcard);
             if (rx.exactMatch(url.toString())) {
-#ifndef NDEBUG
-                // qCDebug(LOG_PLASMA) << md.name() << "matches" << glob << url;
-#endif
                 filtered << md;
             }
         }
@@ -759,9 +748,9 @@ QVector<KPluginMetaData> PluginLoader::listDataEngineMetaData(const QString &par
 
     QVector<KPluginMetaData> plugins;
     if (parentApp.isEmpty()) {
-        plugins = KPluginLoader::findPlugins(PluginLoaderPrivate::s_dataEnginePluginDir);
+        plugins = KPluginMetaData::findPlugins(PluginLoaderPrivate::s_dataEnginePluginDir);
     } else {
-        plugins = KPluginLoader::findPlugins(PluginLoaderPrivate::s_dataEnginePluginDir, filter);
+        plugins = KPluginMetaData::findPlugins(PluginLoaderPrivate::s_dataEnginePluginDir, filter);
     }
 
     return plugins;
@@ -782,16 +771,16 @@ QVector<KPluginMetaData> PluginLoader::listContainmentActionsMetaData(const QStr
 
     QVector<KPluginMetaData> plugins;
     if (parentApp.isEmpty()) {
-        plugins = KPluginLoader::findPlugins(PluginLoaderPrivate::s_containmentActionsPluginDir);
+        plugins = KPluginMetaData::findPlugins(PluginLoaderPrivate::s_containmentActionsPluginDir);
     } else {
-        plugins = KPluginLoader::findPlugins(PluginLoaderPrivate::s_containmentActionsPluginDir, filter);
+        plugins = KPluginMetaData::findPlugins(PluginLoaderPrivate::s_containmentActionsPluginDir, filter);
     }
 
 #if KSERVICE_BUILD_DEPRECATED_SINCE(5, 0)
     // FIXME: this is only for backwards compatibility, but probably will have to stay
     // for the time being
     QSet<QString> knownPlugins;
-    for (const KPluginMetaData &p : qAsConst(plugins)) {
+    for (const KPluginMetaData &p : std::as_const(plugins)) {
         knownPlugins.insert(p.pluginId());
     }
     QString constraint;
@@ -813,6 +802,7 @@ QVector<KPluginMetaData> PluginLoader::listContainmentActionsMetaData(const QStr
     return plugins;
 }
 
+#if PLASMA_BUILD_DEPRECATED_SINCE(5, 86)
 Applet *PluginLoader::internalLoadApplet(const QString &name, uint appletId, const QVariantList &args)
 {
     Q_UNUSED(name)
@@ -872,6 +862,7 @@ KPluginInfo::List PluginLoader::internalContainmentActionsInfo() const
 {
     return KPluginInfo::List();
 }
+#endif
 
 static KPluginInfo::List standardInternalInfo(const QString &type, const QString &category = QString())
 {
@@ -910,7 +901,7 @@ KPluginInfo::List PluginLoader::standardInternalServiceInfo() const
     return standardInternalInfo(QStringLiteral("services"));
 }
 
-QVector<KPluginMetaData> PluginLoaderPrivate::Cache::findPluginsById(const QString &name, const QStringList &dirs)
+KPluginMetaData PluginLoaderPrivate::Cache::findPluginById(const QString &name, const QString &pluginNamespace)
 {
     const qint64 now = qRound64(QDateTime::currentMSecsSinceEpoch() / 1000.0);
     bool useRuntimeCache = true;
@@ -919,18 +910,10 @@ QVector<KPluginMetaData> PluginLoaderPrivate::Cache::findPluginsById(const QStri
         // Find all the plugins now, but only once
         pluginCacheAge = now;
 
-        auto insertIntoCache = [this](const QString &pluginPath) {
-            KPluginMetaData metadata(pluginPath);
-            if (!metadata.isValid()) {
-                qCDebug(LOG_PLASMA) << "invalid metadata" << pluginPath;
-                return;
-            }
-
-            plugins[metadata.pluginId()].append(metadata);
-        };
-
-        for (const QString &dir : dirs)
-            KPluginLoader::forEachPlugin(dir, insertIntoCache);
+        const auto metaDataList = KPluginMetaData::findPlugins(pluginNamespace);
+        for (const KPluginMetaData &metadata : metaDataList) {
+            plugins.insert(metadata.pluginId(), metadata);
+        }
     } else if (now - pluginCacheAge > maxCacheAge) {
         // cache is old and we're not within a few seconds of startup anymore
         useRuntimeCache = false;
@@ -940,22 +923,13 @@ QVector<KPluginMetaData> PluginLoaderPrivate::Cache::findPluginsById(const QStri
     // if name wasn't a path, pluginName == name
     const QString pluginName = name.section(QLatin1Char('/'), -1);
 
-    QVector<KPluginMetaData> ret;
-
     if (useRuntimeCache) {
-        auto it = plugins.constFind(pluginName);
-        if (it != plugins.constEnd()) {
-            ret = *it;
-        }
-        qCDebug(LOG_PLASMA) << "loading applet by name" << name << useRuntimeCache << ret.size();
+        KPluginMetaData data = plugins.value(name);
+        qCDebug(LOG_PLASMA) << "loading applet by name" << name << useRuntimeCache << data.isValid();
+        return data;
     } else {
-        for (const auto &dir : dirs) {
-            ret = KPluginLoader::findPluginsById(dir, pluginName);
-            if (!ret.isEmpty())
-                break;
-        }
+        return KPluginMetaData::findPluginById(pluginNamespace, pluginName);
     }
-    return ret;
 }
 
 } // Plasma Namespace
